@@ -166,7 +166,8 @@ create_context (ctrl_t ctrl, assuan_context_t *r_ctx)
   err = start_new_dirmngr (&ctx,
                            GPG_ERR_SOURCE_DEFAULT,
                            opt.dirmngr_program,
-                           opt.autostart, opt.verbose, DBG_IPC,
+                           opt.autostart?ASSHELP_FLAG_AUTOSTART:0,
+                           opt.verbose, DBG_IPC,
                            NULL /*gpg_status2*/, ctrl);
   if (!opt.autostart && gpg_err_code (err) == GPG_ERR_NO_DIRMNGR)
     {
@@ -389,7 +390,7 @@ ks_status_cb (void *opaque, const char *line)
                 {
                   /* This is an LDAP config entry like
                    * "foo:389:user:pass:base:flags"
-                   * we strip off everything beyound the port.  */
+                   * we strip off everything beyond the port.  */
                   if ((p = strchr (p+1, ':')))
                     {
                       if (p[-1] == ':')
