@@ -41,6 +41,10 @@
 /* Number of bits we accept when reading or writing MPIs. */
 #define MAX_EXTERN_MPI_BITS 16384
 
+/* Number of bytes we accept when reading four-octet count prefixed
+ * key parameters.  Needs to fit as a positive number into an int. */
+#define MAX_EXTERN_KEYPARM_BITS (32768*8)
+
 /* The maximum length of a binary fingerprints.  This is used to
  * provide a static buffer and will be increased if we need to support
  * longer fingerprints.  Warning: At some places we have some
@@ -110,6 +114,14 @@ struct server_control_s
 
   /* This is used to cache a key data base handle.  */
   KEYDB_HANDLE cached_getkey_kdb;
+
+  /* Cached results from HAVEKEY --list.  They are used if the pointer
+   * is not NULL.  The length gives the length in bytes and is a
+   * multiple of 20.  If the no_more flag is set the list shall not
+   * anymore be refreshed even if it has been freed and NULLed. */
+  unsigned char *secret_keygrips;
+  size_t secret_keygrips_len;
+  int no_more_secret_keygrips;
 };
 
 
