@@ -88,6 +88,17 @@ check_key_verify_compliance (PKT_public_key *pk)
 }
 
 
+
+/* Check a signature.  This is shorthand for check_signature2 with
+   the unnamed arguments passed as NULL.  */
+int
+check_signature (ctrl_t ctrl, PKT_signature *sig, gcry_md_hd_t digest)
+{
+  return check_signature2 (ctrl, sig, digest, NULL, 0, NULL,
+                           NULL, NULL, NULL, NULL);
+}
+
+
 /* Check a signature.
  *
  * Looks up the public key that created the signature (SIG->KEYID)
@@ -131,19 +142,14 @@ check_key_verify_compliance (PKT_public_key *pk)
  * If R_PK is not NULL, the public key is stored at that address if it
  * was found; other wise NULL is stored.
  *
- * If R_KEYBLOCK is not NULL, the entire keyblock used to verify the
- * signature is stored at that address.  If no key was found or on
- * some other errors NULL is stored there.  The callers needs to
- * release the keyblock using release_kbnode (kb).
- *
  * Returns 0 on success.  An error code otherwise.  */
 gpg_error_t
-check_signature (ctrl_t ctrl,
-                 PKT_signature *sig, gcry_md_hd_t digest,
-                 const void *extrahash, size_t extrahashlen,
-                 PKT_public_key *forced_pk,
-                 u32 *r_expiredate, int *r_expired, int *r_revoked,
-                 PKT_public_key **r_pk, kbnode_t *r_keyblock)
+check_signature2 (ctrl_t ctrl,
+                  PKT_signature *sig, gcry_md_hd_t digest,
+                  const void *extrahash, size_t extrahashlen,
+                  PKT_public_key *forced_pk,
+                  u32 *r_expiredate,
+		  int *r_expired, int *r_revoked, PKT_public_key **r_pk)
 {
   int rc=0;
   PKT_public_key *pk;
@@ -156,8 +162,6 @@ check_signature (ctrl_t ctrl,
     *r_revoked = 0;
   if (r_pk)
     *r_pk = NULL;
-  if (r_keyblock)
-    *r_keyblock = NULL;
 
   pk = xtrycalloc (1, sizeof *pk);
   if (!pk)
@@ -188,7 +192,7 @@ check_signature (ctrl_t ctrl,
       log_info(_("WARNING: signature digest conflict in message\n"));
       rc = gpg_error (GPG_ERR_GENERAL);
     }
-  else if (get_pubkey_for_sig (ctrl, pk, sig, forced_pk, r_keyblock))
+  else if (get_pubkey_for_sig (ctrl, pk, sig, forced_pk))
     rc = gpg_error (GPG_ERR_NO_PUBKEY);
   else if ((rc = check_key_verify_compliance (pk)))
     ;/* Compliance failure.  */
@@ -787,9 +791,9 @@ check_revocation_keys (ctrl_t ctrl, PKT_public_key *pk, PKT_signature *sig)
           keyid_from_fingerprint (ctrl, pk->revkey[i].fpr, pk->revkey[i].fprlen,
                                   keyid);
 
-          /* If the signature was generated by a designated revoker
-           * verify the signature.  */
-          if (keyid[0] == sig->keyid[0] && keyid[1] == sig->keyid[1])
+          if(keyid[0]==sig->keyid[0] && keyid[1]==sig->keyid[1])
+	    /* The signature was generated by a designated revoker.
+	       Verify the signature.  */
 	    {
               gcry_md_hd_t md;
 
@@ -797,9 +801,8 @@ check_revocation_keys (ctrl_t ctrl, PKT_public_key *pk, PKT_signature *sig)
                 BUG ();
               hash_public_key(md,pk);
 	      /* Note: check_signature only checks that the signature
-	       * is good.  It does not fail if the key is revoked.  */
-              rc = check_signature (ctrl, sig, md, NULL, 0, NULL,
-                                    NULL, NULL, NULL, NULL, NULL);
+		 is good.  It does not fail if the key is revoked.  */
+              rc = check_signature (ctrl, sig, md);
 	      cache_sig_result(sig,rc);
               gcry_md_close (md);
 	      break;
@@ -989,7 +992,7 @@ check_signature_over_key_or_uid (ctrl_t ctrl, PKT_public_key *signer,
               if (ret_pk)
                 {
                   signer = ret_pk;
-                  /* FIXME: Using memset here is probematic because it
+                  /* FIXME: Using memset here is problematic because it
                    * assumes that there are no allocated fields in
                    * SIGNER.  */
                   memset (signer, 0, sizeof (*signer));
@@ -1004,11 +1007,10 @@ check_signature_over_key_or_uid (ctrl_t ctrl, PKT_public_key *signer,
               if (IS_CERT (sig))
                 signer->req_usage = PUBKEY_USAGE_CERT;
 
-              rc = get_pubkey_for_sig (ctrl, signer, sig, NULL, NULL);
+              rc = get_pubkey_for_sig (ctrl, signer, sig, NULL);
               if (rc)
                 {
-                  if (signer_alloced != 1)
-                    xfree (signer);
+                  xfree (signer);
                   signer = NULL;
                   signer_alloced = 0;
                   goto leave;
