@@ -49,7 +49,12 @@ struct mrsw_lock
   int num_readers_active;
   int num_writers_waiting;
   int writer_active;
-  npth_cond_t notify_cond;
+#ifdef HAVE_W32_SYSTEM
+  HANDLE events[2];
+  HANDLE the_event;
+#else
+  int notify_pipe[2];
+#endif
 };
 
 /* MRSW lock to protect the list of cards.
@@ -112,6 +117,7 @@ strcardtype (cardtype_t t)
     case CARDTYPE_GNUK:        return "gnuk";
     case CARDTYPE_YUBIKEY:     return "yubikey";
     case CARDTYPE_ZEITCONTROL: return "zeitcontrol";
+    case CARDTYPE_SCE7:        return "smartcafe";
     }
   return "?";
 }
@@ -365,17 +371,79 @@ card_list_w_unlock (void)
 static void
 card_list_signal (void)
 {
-  npth_cond_broadcast (&card_list_lock.notify_cond);
+#ifdef HAVE_W32_SYSTEM
+  if (SetEvent (card_list_lock.the_event) == 0)
+    log_error ("SetEvent for card_list_signal failed: %s\n",
+               w32_strerror (-1));
+#else
+  write (card_list_lock.notify_pipe[1], "", 1);
+#endif
 }
 
-static void
-card_list_wait (void)
+/* Wait for some card list event or input shutdown.
+   Return 1 on input shutdown, return 0 otherwise.  */
+static int
+card_list_wait (ctrl_t ctrl)
 {
+  gnupg_fd_t fd = ctrl->thread_startup.fd;
+  int nfd;
+  fd_set fdset;
+#ifdef HAVE_W32_SYSTEM
+  unsigned int events_set;
+#endif
+  int ret;
+
   npth_mutex_lock (&card_list_lock.lock);
   card_list_lock.writer_active--;
   npth_cond_broadcast (&card_list_lock.cond);
 
-  npth_cond_wait (&card_list_lock.notify_cond, &card_list_lock.lock);
+  npth_mutex_unlock (&card_list_lock.lock);
+
+  while (1)
+    {
+      FD_ZERO (&fdset);
+      FD_SET (FD2INT (fd), &fdset);
+      nfd = FD2NUM (fd);
+
+#ifdef HAVE_W32_SYSTEM
+      ret = npth_eselect (nfd+1, &fdset, NULL, NULL, NULL,
+                          card_list_lock.events, &events_set);
+#else
+      FD_SET (card_list_lock.notify_pipe[0], &fdset);
+      if (nfd < card_list_lock.notify_pipe[0])
+        nfd = card_list_lock.notify_pipe[0];
+
+      ret = npth_pselect (nfd+1, &fdset, NULL, NULL, NULL,
+                          npth_sigev_sigmask ());
+#endif
+      if (ret <= 0)
+        continue;
+
+      if (FD_ISSET (fd, &fdset))
+        {
+          ret = 1;
+          break;
+        }
+
+#ifdef HAVE_W32_SYSTEM
+      if (events_set & 1)
+        {
+          ret = 0;
+          break;
+        }
+#else
+      if (FD_ISSET (card_list_lock.notify_pipe[0], &fdset))
+        {
+          char buf[256];
+
+          read (card_list_lock.notify_pipe[0], buf, sizeof buf);
+          ret = 0;
+          break;
+        }
+#endif
+    }
+
+  npth_mutex_lock (&card_list_lock.lock);
 
   card_list_lock.num_writers_waiting++;
   while (card_list_lock.num_readers_active
@@ -385,6 +453,7 @@ card_list_wait (void)
 
   card_list_lock.writer_active++;
   npth_mutex_unlock (&card_list_lock.lock);
+  return ret;
 }
 
 
@@ -415,7 +484,7 @@ app_dump_state (void)
  *
  * With KEEP_LOOPING=0, it only outputs once.
  * With KEEP_LOOPING<0, it keeps looping, until it detects no device.
- * With KEEP_LOOPING>0, it keeps looping forever.
+ * With KEEP_LOOPING>0, it keeps looping forever (until connection close).
  */
 gpg_error_t
 app_send_devinfo (ctrl_t ctrl, int keep_looping)
@@ -424,6 +493,13 @@ app_send_devinfo (ctrl_t ctrl, int keep_looping)
   app_t a;
   int no_device;
 
+  /* The connection from client should be by a socket.  This is needed
+     for Windows using the select function.  And it's not good to use
+     the primary pipe connection of gpg-agent for watching
+     devinfo.  */
+  if (keep_looping && ctrl->thread_startup.fd == GNUPG_INVALID_FD)
+    return gpg_error (GPG_ERR_INV_HANDLE);
+
   card_list_w_lock ();
   while (1)
     {
@@ -450,7 +526,8 @@ app_send_devinfo (ctrl_t ctrl, int keep_looping)
       if (keep_looping == 0)
         break;
 
-      card_list_wait ();
+      if (card_list_wait (ctrl))
+        break;
     }
   card_list_w_unlock ();
 
@@ -549,6 +626,51 @@ card_reset (card_t card)
   return err;
 }
 
+
+/* Return the card type from (ATR,ATRLEN) or CARDTYPE_GENERIC in case
+ * of error or if the ATR was not found.  If ATR is NULL, SLOT is used
+ * to retrieve the ATR from the reader.  */
+static cardtype_t
+atr_to_cardtype (int slot, const unsigned char *atr, size_t atrlen)
+{
+#define X(a) ((unsigned char const *)(a))
+  static struct
+  {
+    size_t atrlen;
+    unsigned char const *atr;
+    cardtype_t type;
+  } atrlist[] = {
+    { 19, X("\x3b\xf9\x96\x00\x00\x80\x31\xfe"
+            "\x45\x53\x43\x45\x37\x20\x0f\x00\x20\x46\x4e"),
+      CARDTYPE_SCE7 },
+    { 0 }
+  };
+#undef X
+  unsigned char *atrbuf = NULL;
+  cardtype_t cardtype = 0;
+  int i;
+
+  if (atr)
+    {
+      atrbuf = apdu_get_atr (slot, &atrlen);
+      if (!atrbuf)
+        return 0;
+      atr = atrbuf;
+    }
+
+  for (i=0; atrlist[i].atrlen; i++)
+    if (atrlist[i].atrlen == atrlen
+        && !memcmp (atrlist[i].atr, atr, atrlen))
+      {
+        cardtype = atrlist[i].type;
+        break;
+      }
+  xfree (atrbuf);
+  return cardtype;
+}
+
+
+
 static gpg_error_t
 app_new_register (int slot, ctrl_t ctrl, const char *name,
                   int periodical_check_needed)
@@ -666,13 +788,16 @@ app_new_register (int slot, ctrl_t ctrl, const char *name,
                 }
               xfree (buf);
             }
+          else
+            card->cardtype = atr_to_cardtype (slot, NULL, 0);
         }
-      else
+      else  /* Got 3F00 */
         {
           unsigned char *atr;
           size_t atrlen;
 
           /* This is heuristics to identify different implementations.  */
+          /* FIXME: The first two checks are pretty OpenPGP card specific. */
           atr = apdu_get_atr (slot, &atrlen);
           if (atr)
             {
@@ -680,6 +805,8 @@ app_new_register (int slot, ctrl_t ctrl, const char *name,
                 card->cardtype = CARDTYPE_GNUK;
               else if (atrlen == 21 && atr[7] == 0x75)
                 card->cardtype = CARDTYPE_ZEITCONTROL;
+              else
+                card->cardtype = atr_to_cardtype (slot, atr, atrlen);
               xfree (atr);
             }
         }
@@ -1348,7 +1475,7 @@ app_munge_serialno (card_t card)
        *   !     !----------- Form factor
        *   !----------------- Our prefix
        * Yubico seems to use the decimalized version of their S/N
-       * as the OpenPGP card S/N.  Thus in theory we can contruct the
+       * as the OpenPGP card S/N.  Thus in theory we can construct the
        * number from this information so that we do not rely on having
        * the OpenPGP app enabled.
        */
@@ -2332,6 +2459,15 @@ app_check_pin (card_t card, ctrl_t ctrl, const char *keyidstr,
   return err;
 }
 
+#ifndef HAVE_W32_SYSTEM
+static void
+setup_env (void *arg)
+{
+  char *v = arg;
+
+  putenv (v);
+}
+#endif
 
 static void
 report_change (int slot, int old_status, int cur_status)
@@ -2360,11 +2496,9 @@ report_change (int slot, int old_status, int cur_status)
   else
     {
       gpg_error_t err;
-      const char *args[9], *envs[2];
+      const char *args[9];
       char numbuf1[30], numbuf2[30], numbuf3[30];
-
-      envs[0] = envstr;
-      envs[1] = NULL;
+      gnupg_spawn_actions_t act = NULL;
 
       sprintf (numbuf1, "%d", slot);
       sprintf (numbuf2, "0x%04X", old_status);
@@ -2382,7 +2516,16 @@ report_change (int slot, int old_status, int cur_status)
       args[8] = NULL;
 
       fname = make_filename (gnupg_homedir (), "scd-event", NULL);
-      err = gnupg_spawn_process_detached (fname, args, envs);
+      err = gnupg_spawn_actions_new (&act);
+      if (!err)
+        {
+#ifndef HAVE_W32_SYSTEM
+          gnupg_spawn_actions_set_atfork (act, setup_env, envstr);
+#endif
+          err = gnupg_process_spawn (fname, args, GNUPG_PROCESS_DETACHED,
+                                     act, NULL);
+          gnupg_spawn_actions_release (act);
+        }
       if (err && gpg_err_code (err) != GPG_ERR_ENOENT)
         log_error ("failed to run event handler '%s': %s\n",
                    fname, gpg_strerror (err));
@@ -2479,6 +2622,9 @@ gpg_error_t
 initialize_module_command (void)
 {
   gpg_error_t err;
+#ifndef HAVE_W32_SYSTEM
+  int ret;
+#endif
 
   card_list_lock.num_readers_active = 0;
   card_list_lock.num_writers_waiting = 0;
@@ -2499,13 +2645,17 @@ initialize_module_command (void)
       return err;
     }
 
-  err = npth_cond_init (&card_list_lock.notify_cond, NULL);
-  if (err)
+#ifdef HAVE_W32_SYSTEM
+  scd_init_event (&card_list_lock.the_event, card_list_lock.events);
+#else
+  ret = gnupg_create_pipe (card_list_lock.notify_pipe);
+  if (ret)
     {
       err = gpg_error_from_syserror ();
-      log_error ("npth_cond_init failed: %s\n", gpg_strerror (err));
+      log_error ("pipe creation failed: %s\n", gpg_strerror (ret));
       return err;
     }
+#endif
 
   return apdu_init ();
 }
