@@ -151,7 +151,7 @@ hex_to_buffer (const char *string, size_t *r_length)
 /* Reset the card and free the application context.  With SEND_RESET
    set to true actually send a RESET to the reader; this is the normal
    way of calling the function.  If KEEP_LOCK is set and the session
-   is locked that lock wil not be released.  */
+   is locked that lock will not be released.  */
 static void
 do_reset (ctrl_t ctrl, int send_reset, int keep_lock)
 {
@@ -957,7 +957,8 @@ pin_cb (void *opaque, const char *info, char **retstr)
          We ignore any value returned.  */
       if (info)
         {
-          log_debug ("prompting for pinpad entry '%s'\n", info);
+          if (DBG_IPC)
+            log_debug ("prompting for pinpad entry '%s'\n", info);
           rc = gpgrt_asprintf (&command, "POPUPPINPADPROMPT %s", info);
           if (rc < 0)
             return gpg_error (gpg_err_code_from_errno (errno));
@@ -966,7 +967,8 @@ pin_cb (void *opaque, const char *info, char **retstr)
         }
       else
         {
-          log_debug ("dismiss pinpad entry prompt\n");
+          if (DBG_IPC)
+            log_debug ("dismiss pinpad entry prompt\n");
           rc = assuan_inquire (ctx, "DISMISSPINPADPROMPT",
                                &value, &valuelen, MAXLEN_PIN);
         }
@@ -976,7 +978,8 @@ pin_cb (void *opaque, const char *info, char **retstr)
     }
 
   *retstr = NULL;
-  log_debug ("asking for PIN '%s'\n", info);
+  if (DBG_IPC)
+    log_debug ("asking for PIN '%s'\n", info);
 
   rc = gpgrt_asprintf (&command, "NEEDPIN %s", info);
   if (rc < 0)
@@ -1353,7 +1356,7 @@ static const char hlp_writecert[] =
   "\n"
   "This command is used to store a certificate on a smartcard.  The\n"
   "allowed certids depend on the currently selected smartcard\n"
-  "application. The actual certifciate is requested using the inquiry\n"
+  "application. The actual certificate is requested using the inquiry\n"
   "\"CERTDATA\" and needs to be provided in its raw (e.g. DER) form.\n"
   "\n"
   "In almost all cases a PIN will be requested.  See the related\n"
@@ -2472,9 +2475,8 @@ register_commands (assuan_context_t ctx)
 
 
 /* Startup the server.  If FD is given as -1 this is simple pipe
-   server, otherwise it is a regular server.  Returns true if there
-   are no more active asessions.  */
-int
+   server, otherwise it is a regular server.  */
+void
 scd_command_handler (ctrl_t ctrl, gnupg_fd_t fd)
 {
   int rc;
@@ -2573,9 +2575,6 @@ scd_command_handler (ctrl_t ctrl, gnupg_fd_t fd)
 
   if (stopme)
     scd_exit (0);
-
-  /* If there are no more sessions return true.  */
-  return !session_list;
 }
 
 
@@ -2942,10 +2941,10 @@ void
 send_client_notifications (card_t card, int removal)
 {
   struct {
-    pid_t pid;
 #ifdef HAVE_W32_SYSTEM
     HANDLE handle;
 #else
+    pid_t pid;
     int signo;
 #endif
   } killed[50];
@@ -2965,10 +2964,10 @@ send_client_notifications (card_t card, int removal)
 
       if (sl->ctrl_backlink && sl->ctrl_backlink->card_ctx == card)
         {
-          pid_t pid;
 #ifdef HAVE_W32_SYSTEM
           HANDLE handle;
 #else
+          pid_t pid;
           int signo;
 #endif
 
@@ -2983,32 +2982,33 @@ send_client_notifications (card_t card, int removal)
           if (!sl->event_signal || !sl->assuan_ctx)
             continue;
 
-          pid = assuan_get_pid (sl->assuan_ctx);
-
 #ifdef HAVE_W32_SYSTEM
           handle = sl->event_signal;
           for (kidx=0; kidx < killidx; kidx++)
-            if (killed[kidx].pid == pid
-                && killed[kidx].handle == handle)
+            if (killed[kidx].handle == handle)
               break;
           if (kidx < killidx)
-            log_info ("event %p (%p) already triggered for client %d\n",
-                      sl->event_signal, handle, (int)pid);
+            {
+              if (opt.verbose)
+                log_info ("event %p already triggered for client\n",
+                          sl->event_signal);
+            }
           else
             {
-              log_info ("triggering event %p (%p) for client %d\n",
-                        sl->event_signal, handle, (int)pid);
+              if (opt.verbose)
+                log_info ("triggering event %p for client\n",
+                          sl->event_signal);
               if (!SetEvent (handle))
                 log_error ("SetEvent(%p) failed: %s\n",
                            sl->event_signal, w32_strerror (-1));
               if (killidx < DIM (killed))
                 {
-                  killed[killidx].pid = pid;
                   killed[killidx].handle = handle;
                   killidx++;
                 }
             }
 #else /*!HAVE_W32_SYSTEM*/
+          pid = assuan_get_pid (sl->assuan_ctx);
           signo = sl->event_signal;
 
           if (pid != (pid_t)(-1) && pid && signo > 0)
@@ -3018,12 +3018,16 @@ send_client_notifications (card_t card, int removal)
                     && killed[kidx].signo == signo)
                   break;
               if (kidx < killidx)
-                log_info ("signal %d already sent to client %d\n",
-                          signo, (int)pid);
+                {
+                  if (opt.verbose)
+                    log_info ("signal %d already sent to client %d\n",
+                              signo, (int)pid);
+                }
               else
                 {
-                  log_info ("sending signal %d to client %d\n",
-                            signo, (int)pid);
+                  if (opt.verbose)
+                    log_info ("sending signal %d to client %d\n",
+                              signo, (int)pid);
                   kill (pid, signo);
                   if (killidx < DIM (killed))
                     {
