@@ -281,9 +281,6 @@ static gnupg_fd_t create_server_socket (const char *name,
 static void *start_connection_thread (void *arg);
 static void handle_connections (gnupg_fd_t listen_fd);
 
-/* Pth wrapper function definitions. */
-ASSUAN_SYSTEM_NPTH_IMPL;
-
 static int active_connections;
 
 
@@ -489,7 +486,6 @@ main (int argc, char **argv )
   malloc_hooks.free = gcry_free;
   assuan_set_malloc_hooks (&malloc_hooks);
   assuan_set_gpg_err_source (GPG_ERR_SOURCE_DEFAULT);
-  assuan_set_system_hooks (ASSUAN_SYSTEM_NPTH);
   assuan_sock_init ();
   setup_libassuan_logging (&opt.debug, NULL);
 
@@ -536,12 +532,12 @@ main (int argc, char **argv )
      Now we are working under our real uid
   */
 
-  /* The configuraton directories for use by gpgrt_argparser.  */
+  /* The configuration directories for use by gpgrt_argparser.  */
   gpgrt_set_confdir (GPGRT_CONFDIR_SYS, gnupg_sysconfdir ());
   gpgrt_set_confdir (GPGRT_CONFDIR_USER, gnupg_homedir ());
 
   /* We are re-using the struct, thus the reset flag.  We OR the
-   * flags so that the internal intialized flag won't be cleared. */
+   * flags so that the internal initialized flag won't be cleared. */
   argc = orig_argc;
   argv = orig_argv;
   pargs.argc = &argc;
@@ -780,6 +776,7 @@ main (int argc, char **argv )
       npth_init ();
       setup_signal_mask ();
       gpgrt_set_syscall_clamp (npth_unprotect, npth_protect);
+      assuan_control (ASSUAN_CONTROL_REINIT_SYSCALL_CLAMP, NULL);
 
       /* If --debug-allow-core-dump has been given we also need to
          switch the working directory to a place where we can actually
@@ -921,6 +918,7 @@ main (int argc, char **argv )
       npth_init ();
       setup_signal_mask ();
       gpgrt_set_syscall_clamp (npth_unprotect, npth_protect);
+      assuan_control (ASSUAN_CONTROL_REINIT_SYSCALL_CLAMP, NULL);
 
       /* Detach from tty and put process into a new session. */
       if (!nodetach )
@@ -1234,7 +1232,7 @@ start_connection_thread (void *arg)
       && assuan_sock_check_nonce (ctrl->thread_startup.fd, &socket_nonce))
     {
       log_info (_("error reading nonce on fd %d: %s\n"),
-                FD2INT(ctrl->thread_startup.fd), strerror (errno));
+                FD_DBG (ctrl->thread_startup.fd), strerror (errno));
       assuan_sock_close (ctrl->thread_startup.fd);
       xfree (ctrl);
       return NULL;
@@ -1245,7 +1243,7 @@ start_connection_thread (void *arg)
   scd_init_default_ctrl (ctrl);
   if (opt.verbose)
     log_info (_("handler for fd %d started\n"),
-              FD2INT(ctrl->thread_startup.fd));
+              FD_DBG (ctrl->thread_startup.fd));
 
   scd_command_handler (ctrl, ctrl->thread_startup.fd);
 
@@ -1258,7 +1256,7 @@ start_connection_thread (void *arg)
 
   if (opt.verbose)
     log_info (_("handler for fd %d terminated\n"),
-              FD2INT (ctrl->thread_startup.fd));
+              FD_DBG (ctrl->thread_startup.fd));
 
   scd_deinit_default_ctrl (ctrl);
   xfree (ctrl);
@@ -1341,7 +1339,7 @@ handle_connections (gnupg_fd_t listen_fd)
   if (listen_fd != GNUPG_INVALID_FD)
     {
       FD_SET (FD2INT (listen_fd), &fdset);
-      nfd = FD2INT (listen_fd);
+      nfd = FD2NUM (listen_fd);
     }
 
   for (;;)
@@ -1427,8 +1425,8 @@ handle_connections (gnupg_fd_t listen_fd)
           gnupg_fd_t fd;
 
           plen = sizeof paddr;
-          fd = INT2FD (npth_accept (FD2INT (listen_fd),
-                                    (struct sockaddr *)&paddr, &plen));
+          fd = assuan_sock_accept (listen_fd,
+                                   (struct sockaddr *)&paddr, &plen);
           if (fd == GNUPG_INVALID_FD)
             {
               log_error ("accept failed: %s\n", strerror (errno));
@@ -1445,7 +1443,7 @@ handle_connections (gnupg_fd_t listen_fd)
               npth_t thread;
 
               snprintf (threadname, sizeof threadname, "conn fd=%d",
-                        FD2INT (fd));
+                        FD_DBG (fd));
               ctrl->thread_startup.fd = fd;
               ret = npth_create (&thread, &tattr, start_connection_thread, ctrl);
               if (ret)
