@@ -181,10 +181,6 @@ struct
    * attribute values.  */
   strlist_t attributes;
 
-  /* The list of --assert-signer option values.  Note: The values are
-   * modified to uppercase if they represent a fingerrint */
-  strlist_t assert_signer_list;
-
   /* Compatibility flags (COMPAT_FLAG_xxxx).  */
   unsigned int compat_flags;
 } opt;
@@ -308,7 +304,7 @@ struct rootca_flags_s
                             information.  */
   unsigned int relax:1;  /* Relax checking of root certificates.  */
   unsigned int chain_model:1; /* Root requires the use of the chain model.  */
-  unsigned int qualified:1;   /* Root CA used for qualfied signatures.   */
+  unsigned int qualified:1;   /* Root CA used for qualified signatures.   */
   unsigned int de_vs:1;       /* Root CA is de-vs compliant.             */
 };
 
@@ -316,7 +312,6 @@ struct rootca_flags_s
 
 /*-- gpgsm.c --*/
 extern int gpgsm_errors_seen;
-extern int assert_signer_true;
 
 void gpgsm_exit (int rc);
 void gpgsm_init_default_ctrl (struct server_control_s *ctrl);
@@ -436,9 +431,9 @@ gpg_error_t gpgsm_show_certs (ctrl_t ctrl, int nfiles, char **files,
                               estream_t fp);
 
 /*-- import.c --*/
-int gpgsm_import (ctrl_t ctrl, int in_fd, int reimport_mode);
+int gpgsm_import (ctrl_t ctrl, estream_t in_fp, int reimport_mode);
 int gpgsm_import_files (ctrl_t ctrl, int nfiles, char **files,
-                        int (*of)(const char *fname));
+                        estream_t (*of)(const char *fname, const char *mode));
 
 /*-- export.c --*/
 void gpgsm_export (ctrl_t ctrl, strlist_t names, estream_t stream);
@@ -449,23 +444,24 @@ void gpgsm_p12_export (ctrl_t ctrl, const char *name, estream_t stream,
 int gpgsm_delete (ctrl_t ctrl, strlist_t names);
 
 /*-- verify.c --*/
-int gpgsm_verify (ctrl_t ctrl, int in_fd, int data_fd, estream_t out_fp);
+int gpgsm_verify (ctrl_t ctrl, estream_t in_fp, estream_t data_fp,
+                  estream_t out_fp);
 
 /*-- sign.c --*/
 int gpgsm_get_default_cert (ctrl_t ctrl, ksba_cert_t *r_cert);
 int gpgsm_sign (ctrl_t ctrl, certlist_t signerlist,
-                int data_fd, int detached, estream_t out_fp);
+                estream_t data_fp, int detached, estream_t out_fp);
 
 /*-- encrypt.c --*/
 int gpgsm_encrypt (ctrl_t ctrl, certlist_t recplist,
-                   int in_fd, estream_t out_fp);
+                   estream_t in_fp, estream_t out_fp);
 
 /*-- decrypt.c --*/
 gpg_error_t ecdh_derive_kek (unsigned char *key, unsigned int keylen,
                              int hash_algo, const char *wrap_algo_str,
                              const void *secret, unsigned int secretlen,
                              const void *ukm, unsigned int ukmlen);
-int gpgsm_decrypt (ctrl_t ctrl, int in_fd, estream_t out_fp);
+int gpgsm_decrypt (ctrl_t ctrl, estream_t in_fp, estream_t out_fp);
 
 /*-- certreqgen.c --*/
 int gpgsm_genkey (ctrl_t ctrl, estream_t in_stream, estream_t out_stream);
