@@ -396,7 +396,7 @@ start_agent (assuan_context_t *ctx_p)
                              opt.agent_program,
                              NULL, NULL,
                              session_env,
-                             opt.autostart,
+                             opt.autostart?ASSHELP_FLAG_AUTOSTART:0,
                              !opt.quiet, 0,
                              NULL, NULL);
 
@@ -818,7 +818,6 @@ ssh_authorized_keys (const char *user, struct ssh_key_list **r_ssh_key_list)
       xfree (fname);
       return err;
     }
-  xfree (fname);
 
   maxlen = 2048; /* Set limit.  */
   while ((len = es_read_line (fp, &line, &length_of_line, &maxlen)) > 0)
@@ -861,6 +860,7 @@ ssh_authorized_keys (const char *user, struct ssh_key_list **r_ssh_key_list)
   *r_ssh_key_list = ssh_key_list;
 
  leave:
+  xfree (fname);
   xfree (line);
   es_fclose (fp);
   return err;
